// Copyright (c) 2021 Ron Booth. All rights reserved.
// Use of this source code is governed by a license that can be found in the
// LICENSE file.

import 'dart:collection';
import 'dart:math' as math;
import 'dart:ui' as ui show PlaceholderAlignment, TextBox;

import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

import 'float_tag.dart';
import 'inline_span_ext.dart';
import 'render_float_column.dart';
import 'render_text_mixin.dart';
import 'shared.dart';
import 'wrappable_text.dart';

///
/// WrappableTextRenderer
///
class WrappableTextRenderer {
  WrappableTextRenderer(
    RenderBox parent,
    WrappableText wt,
    TextDirection defaultTextDirection,
    DefaultTextStyle defaultTextStyle,
    double defaultTextScaleFactor,
  ) : renderer = TextRenderer(
            parent,
            TextPainter(
                text: TextSpan(
                    style: defaultTextStyle.style, children: [wt.text]),
                textAlign: wt.textAlign ??
                    defaultTextStyle.textAlign ??
                    TextAlign.start,
                textDirection: wt.textDirection ?? defaultTextDirection,
                textScaleFactor: wt.textScaleFactor ?? defaultTextScaleFactor,
                locale: wt.locale,
                strutStyle: wt.strutStyle,
                textHeightBehavior: wt.textHeightBehavior),
            0);

  final TextRenderer renderer;

  final subs = <TextRenderer>[];

  TextRenderer operator [](int index) => index == -1 ? renderer : subs[index];

  TextDirection get textDirection => renderer._painter.textDirection!;

  List<TextRenderer> get renderers => subs.isNotEmpty ? subs : [renderer];

  TextRenderer rendererWithPlaceholder(int index) {
    if (subs.isEmpty) {
      return renderer;
    } else {
      var i = index;
      for (final sub in subs) {
        final count = sub.placeholderSpans.length;
        if (i < count) {
          return sub;
        } else {
          i -= count;
        }
      }
      assert(false);
      return renderer;
    }
  }

  void updateWith(
    WrappableText wt,
    RenderBox parent,
    TextDirection defaultTextDirection,
    DefaultTextStyle defaultTextStyle,
    double defaultTextScaleFactor,
  ) {
    var needsPaint = false;
    var needsLayout = false;

    final textSpan =
        TextSpan(style: defaultTextStyle.style, children: [wt.text]);
    switch (renderer._painter.text!.compareTo(textSpan)) {
      case RenderComparison.identical:
      case RenderComparison.metadata:
        break;
      case RenderComparison.paint:
        renderer._painter.text = textSpan;
        renderer._cachedAttributedLabel = null;
        renderer._cachedCombinedSemanticsInfos = null;
        renderer.clearPlaceholderSpans();
        needsPaint = true;
        break;
      case RenderComparison.layout:
        renderer._painter.text = textSpan;
        renderer._cachedAttributedLabel = null;
        renderer._cachedCombinedSemanticsInfos = null;
        renderer.clearPlaceholderSpans();
        needsLayout = true;
        break;
    }

    final textAlign =
        wt.textAlign ?? defaultTextStyle.textAlign ?? TextAlign.start;
    if (renderer._painter.textAlign != textAlign) {
      renderer._painter.textAlign = textAlign;
      needsLayout = true;
    }

    final textDirection = wt.textDirection ?? defaultTextDirection;
    if (renderer._painter.textDirection != textDirection) {
      renderer._painter.textDirection = textDirection;
      needsLayout = true;
    }

    final textScaleFactor = wt.textScaleFactor ?? defaultTextScaleFactor;
    if (renderer._painter.textScaleFactor != textScaleFactor) {
      renderer._painter.textScaleFactor = textScaleFactor;
      needsLayout = true;
    }

    if (renderer._painter.locale != wt.locale) {
      renderer._painter.locale = wt.locale;
      needsLayout = true;
    }

    if (renderer._painter.strutStyle != wt.strutStyle) {
      renderer._painter.strutStyle = wt.strutStyle;
      needsLayout = true;
    }

    if (renderer._painter.textHeightBehavior != wt.textHeightBehavior) {
      renderer._painter.textHeightBehavior = wt.textHeightBehavior;
      needsLayout = true;
    }

    if (needsLayout) {
      parent.markNeedsLayout();
    } else if (needsPaint) {
      parent
        ..markNeedsPaint()
        ..markNeedsSemanticsUpdate();
    }
  }
}

///
/// TextRenderer
///
class TextRenderer with RenderTextMixin {
  TextRenderer(this._parent, this._painter, this.startingPlaceholderIndex)
      : assert(_painter.text != null);

  final RenderBox _parent;
  final TextPainter _painter;
  final int startingPlaceholderIndex;
  List<PlaceholderSpan>? _placeholderSpans;

  Offset? _offset;
  set offset(Offset value) => _offset = value;

  int get nextPlaceholderIndex =>
      startingPlaceholderIndex + placeholderSpans.length;

  List<PlaceholderSpan> get placeholderSpans {
    if (_placeholderSpans == null) {
      _placeholderSpans = <PlaceholderSpan>[];
      _painter.text!.visitChildren((span) {
        if (span is PlaceholderSpan) _placeholderSpans!.add(span);
        return true;
      });
    }
    return _placeholderSpans!;
  }

  void clearPlaceholderSpans() => _placeholderSpans = null;

  /// Computes the visual position of the glyphs for painting the text and the
  /// position of the inline widget children.
  void layout(BoxConstraints constraints) {
    _painter.layout(
        minWidth: constraints.minWidth, maxWidth: constraints.maxWidth);
  }

  TextRenderer copyWith(
    InlineSpan text,
    int startingPlaceholderIndex,
  ) =>
      TextRenderer(
          _parent,
          TextPainter(
              text: text,
              textAlign: _painter.textAlign,
              textDirection: _painter.textDirection,
              textScaleFactor: _painter.textScaleFactor,
              locale: _painter.locale,
              strutStyle: _painter.strutStyle,
              textHeightBehavior: _painter.textHeightBehavior),
          startingPlaceholderIndex);

  TextBox placeholderBoxForWidgetIndex(int index) {
    final i = index - startingPlaceholderIndex;
    if ((_painter.inlinePlaceholderBoxes?.length ?? 0) > i) {
      return _painter.inlinePlaceholderBoxes![i];
    } else {
      assert(false);
      return TextBox.fromLTRBD(0, 0, 0, 0, _painter.textDirection!);
    }
  }

  double placeholderScaleForWidgetIndex(int index) {
    final i = index - startingPlaceholderIndex;
    if ((_painter.inlinePlaceholderScales?.length ?? 0) > i) {
      return _painter.inlinePlaceholderScales![i];
    } else {
      assert(false);
      return 1.0;
    }
  }

  /// Returns an estimate of the initial line height based on the initial font
  /// size, initial line height scale, and the text scale factor.
  double initialLineHeight() {
    final fontSize = _painter.text!.initialFontSize(14.0);
    final lineHeightScale = _painter.text!.initialLineHeightScale(1.12);
    return fontSize * lineHeightScale * _painter.textScaleFactor;
  }

  double initialScaledFontSize() {
    final fontSize = _painter.text!.initialFontSize(14.0);
    return fontSize * _painter.textScaleFactor;
  }

  /// Sets the placeholder dimensions for this paragraph's inline widget
  /// children, if any. Returns true iff any of the children are floated.
  bool setPlaceholderDimensions(
    RenderBox? firstChild,
    BoxConstraints constraints,
    double textScaleFactor,
  ) {
    if (firstChild == null) return false;

    final paragraphIndex = firstChild.tag.index;

    // The children will be scaled by textScaleFactor during the painting
    // phase, so the constraints are divided by textScaleFactor.
    final childConstraints =
        BoxConstraints(maxWidth: constraints.maxWidth) / textScaleFactor;

    final placeholderDimensions = List<PlaceholderDimensions>.filled(
        placeholderSpans.length, PlaceholderDimensions.empty,
        growable: false);

    var hasFloatedChildren = false;
    RenderBox? child = firstChild;
    while (child != null && child.tag.index == paragraphIndex) {
      final childParentData = child.parentData! as FloatColumnParentData;

      final i = child.tag.placeholderIndex - startingPlaceholderIndex;
      if (i >= 0 && i < placeholderSpans.length) {
        if (child.tag.float != FCFloat.none) {
          hasFloatedChildren = true;
          if (!child.hasSize) {
            _layoutChild(child, i, BoxConstraints.tight(Size.zero));
          }
        } else {
          placeholderDimensions[i] = _layoutChild(child, i, childConstraints);
        }
      }

      child = childParentData.nextSibling;
    }

    _painter.setPlaceholderDimensions(placeholderDimensions);

    return hasFloatedChildren;
  }

  /// Layout the [child] inline widget at the given [childIndex].
  PlaceholderDimensions _layoutChild(
    RenderBox child,
    int childIndex,
    BoxConstraints constraints, {
    bool dry = false,
  }) {
    assert(childIndex >= 0 && childIndex < placeholderSpans.length);

    double? baselineOffset;
    final Size childSize;
    if (!dry) {
      if (!child.hasSize) {
        // TODO(ron): Maybe need to call this every time in case constraints
        // change?
        child.layout(
          constraints,
          parentUsesSize: true,
        );
      }
      childSize = child.size;
      switch (placeholderSpans[childIndex].alignment) {
        case ui.PlaceholderAlignment.baseline:
          baselineOffset = child
              .getDistanceToBaseline(placeholderSpans[childIndex].baseline!);
          break;
        default:
          baselineOffset = null;
          break;
      }
    } else {
      assert(placeholderSpans[childIndex].alignment !=
          ui.PlaceholderAlignment.baseline);
      childSize = child.getDryLayout(constraints);
    }

    return PlaceholderDimensions(
      size: childSize,
      alignment: placeholderSpans[childIndex].alignment,
      baseline: placeholderSpans[childIndex].baseline,
      baselineOffset: baselineOffset,
    );
  }

  //
  // Semantics related:
  //

  /// Collected during `describeSemanticsConfiguration`, used by
  /// `assembleSemanticsNode`.
  AttributedString? _cachedAttributedLabel;
  List<InlineSpanSemanticsInformation>? _cachedCombinedSemanticsInfos;
/*
  List<InlineSpanSemanticsInformation>? _semanticsInfo;

  void describeSemanticsConfiguration(SemanticsConfiguration config) {
    _semanticsInfo = text.getSemanticsInformation();

    if (_semanticsInfo!.any((info) => info.recognizer != null)) {
      config
        ..explicitChildNodes = true
        ..isSemanticBoundary = true;
    } else {
      if (_cachedAttributedLabel == null) {
        final buffer = StringBuffer();
        var offset = 0;
        final attributes = <StringAttribute>[];
        for (final info in _semanticsInfo!) {
          final label = info.semanticsLabel ?? info.text;
          for (final infoAttribute in info.stringAttributes) {
            final originalRange = infoAttribute.range;
            attributes.add(
              infoAttribute.copy(
                  range: TextRange(
                      start: offset + originalRange.start,
                      end: offset + originalRange.end)),
            );
          }
          buffer.write(label);
          offset += label.length;
        }
        _cachedAttributedLabel =
            AttributedString(buffer.toString(), attributes: attributes);
      }
      config
        ..attributedLabel = _cachedAttributedLabel!
        ..textDirection = textDirection;
    }
  }

  // Caches [SemanticsNode]s created during [assembleSemanticsNode] so they
  // can be re-used when [assembleSemanticsNode] is called again. This ensures
  // stable ids for the [SemanticsNode]s of [TextSpan]s across
  // [assembleSemanticsNode] invocations.
  Queue<SemanticsNode>? _cachedChildNodes;

  void assembleSemanticsNode(
    SemanticsNode node,
    SemanticsConfiguration config,
    Iterable<SemanticsNode> children,
  ) {
    assert(_semanticsInfo != null && _semanticsInfo!.isNotEmpty);
    final newChildren = <SemanticsNode>[];
    var currentDirection = textDirection;
    Rect currentRect;
    var ordinal = 0.0;
    var start = 0;
    var placeholderIndex = 0;
    var childIndex = 0;
    RenderBox? child = firstChild;
    final newChildCache = Queue<SemanticsNode>();
    _cachedCombinedSemanticsInfos ??= combineSemanticsInfo(_semanticsInfo!);
    for (final info in _cachedCombinedSemanticsInfos!) {
      final selection = TextSelection(
        baseOffset: start,
        extentOffset: start + info.text.length,
      );
      start += info.text.length;

      if (info.isPlaceholder) {
        // A placeholder span may have 0 to multiple semantics nodes, we need
        // to annotate all of the semantics nodes that belong to this span.
        while (children.length > childIndex &&
            children
                .elementAt(childIndex)
                .isTagged(PlaceholderSpanIndexSemanticsTag(placeholderIndex))) {
          final childNode = children.elementAt(childIndex);
          final parentData = child!.parentData! as TextParentData;
          childNode.rect = Rect.fromLTWH(
            childNode.rect.left,
            childNode.rect.top,
            childNode.rect.width * parentData.scale!,
            childNode.rect.height * parentData.scale!,
          );
          newChildren.add(childNode);
          childIndex += 1;
        }
        child = childAfter(child!);
        placeholderIndex += 1;
      } else {
        final initialDirection = currentDirection;
        final rects = getBoxesForSelection(selection);
        if (rects.isEmpty) {
          continue;
        }
        var rect = rects.first.toRect();
        currentDirection = rects.first.direction;
        for (final textBox in rects.skip(1)) {
          rect = rect.expandToInclude(textBox.toRect());
          currentDirection = textBox.direction;
        }
        // Any of the text boxes may have had infinite dimensions.
        // We shouldn't pass infinite dimensions up to the bridges.
        rect = Rect.fromLTWH(
          math.max(0.0, rect.left),
          math.max(0.0, rect.top),
          math.min(rect.width, constraints.maxWidth),
          math.min(rect.height, constraints.maxHeight),
        );
        // round the current rectangle to make this API testable and add some
        // padding so that the accessibility rects do not overlap with the text.
        currentRect = Rect.fromLTRB(
          rect.left.floorToDouble() - 4.0,
          rect.top.floorToDouble() - 4.0,
          rect.right.ceilToDouble() + 4.0,
          rect.bottom.ceilToDouble() + 4.0,
        );
        final configuration = SemanticsConfiguration()
          ..sortKey = OrdinalSortKey(ordinal++)
          ..textDirection = initialDirection
          ..attributedLabel = AttributedString(info.semanticsLabel ?? info.text,
              attributes: info.stringAttributes);
        final recognizer = info.recognizer;
        if (recognizer != null) {
          if (recognizer is TapGestureRecognizer) {
            if (recognizer.onTap != null) {
              configuration
                ..onTap = recognizer.onTap
                ..isLink = true;
            }
          } else if (recognizer is DoubleTapGestureRecognizer) {
            if (recognizer.onDoubleTap != null) {
              configuration
                ..onTap = recognizer.onDoubleTap
                ..isLink = true;
            }
          } else if (recognizer is LongPressGestureRecognizer) {
            if (recognizer.onLongPress != null) {
              configuration.onLongPress = recognizer.onLongPress;
            }
          } else {
            assert(false, '${recognizer.runtimeType} is not supported.');
          }
        }
        final newChild = (_cachedChildNodes?.isNotEmpty == true)
            ? _cachedChildNodes!.removeFirst()
            : SemanticsNode();
        // ignore: cascade_invocations
        newChild
          ..updateWith(config: configuration)
          ..rect = currentRect;
        newChildCache.addLast(newChild);
        newChildren.add(newChild);
      }
    }
    // Makes sure we annotated all of the semantics children.
    assert(childIndex == children.length);
    assert(child == null);

    _cachedChildNodes = newChildCache;
    node.updateWith(config: config, childrenInInversePaintOrder: newChildren);
  }

  void clearSemantics() {
    _cachedChildNodes = null;
  }
*/
  //
  // RenderTextAdapter overrides:
  //

  @override
  List<TextBox> getBoxesForSelection(TextSelection selection) =>
      _painter.getBoxesForSelection(selection);

  @override
  double? getFullHeightForCaret(TextPosition position) =>
      _painter.getFullHeightForCaret(position, Rect.zero);

  @override
  Offset getOffsetForCaret(TextPosition position, Rect caretPrototype) =>
      _painter.getOffsetForCaret(position, caretPrototype);

  @override
  TextPosition getPositionForOffset(Offset offset) =>
      _painter.getPositionForOffset(offset);

  @override
  TextRange getWordBoundary(TextPosition position) =>
      _painter.getWordBoundary(position);

  @override
  double get height => _painter.height;

  @override
  Locale? get locale => _painter.locale;

  @override
  int? get maxLines => _painter.maxLines;

  @override
  Offset get offset => _offset!;

  @override
  RenderBox get renderBox => _parent;

  @override
  StrutStyle? get strutStyle => _painter.strutStyle;

  @override
  InlineSpan get text => _painter.text!;

  @override
  TextAlign get textAlign => _painter.textAlign;

  @override
  TextDirection get textDirection => _painter.textDirection!;

  @override
  TextHeightBehavior? get textHeightBehavior => _painter.textHeightBehavior;

  @override
  double get textScaleFactor => _painter.textScaleFactor;

  @override
  Size get textSize => _painter.size;

  @override
  TextWidthBasis get textWidthBasis => _painter.textWidthBasis;

  @override
  void paint(PaintingContext context, Offset offset) {
    _painter.paint(context.canvas, this.offset + offset);
  }

  @override
  double get width => _painter.width;
}

extension on RenderBox {
  FloatTag get tag => ((this as RenderMetaData).metaData as FloatTag);
}
